# Copyright 2021 Richard Maynard (richard.maynard@gmail.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pkg_resources import DistributionNotFound, get_distribution
from tfworker.commands.base import BaseCommand


class VersionCommand(BaseCommand):
    def __init__(self):
        try:
            pkg_info = get_distribution("terraform-worker")
            self._version = pkg_info.version
        except DistributionNotFound:
            self._version = "unknown"

    def exec(self):
        print(f"terraform-worker version {self._version}")



# Copyright 2021 Richard Maynard (richard.maynard@gmail.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.import collections
import os
import re
import shutil
import tempfile
from abc import ABCMeta, abstractmethod, abstractstaticmethod
from pathlib import Path
from typing import Callable

from tfworker.util.system import pipe_exec


class CopyFactory:
    """ The factory class for creating copiers """

    registry = {}

    @classmethod
    def register(cls, name: str) -> Callable:
        """ Class method to register copiers """

        def inner_wrapper(wrapped_class: Copier) -> Callable:
            if name in cls.registry:
                raise ValueError(f"Executor {name} already exists")
            cls.registry[name] = wrapped_class
            return wrapped_class

        return inner_wrapper

    @classmethod
    def create(cls, source: str, **kwargs) -> "Copier":
        """ create returns a copier type that supports handling the provided source """
        copier_class = cls.registry[cls.get_copier_type(source, **kwargs)]
        copier = copier_class(source, **kwargs)
        return copier

    @classmethod
    def get_copier_type(cls, source: str, **kwargs) -> str:
        """ get_copier_type tries to find a supported copier based on the provided source """
        for copier_type, copier_class in cls.registry.items():
            if copier_class.type_match(source, **kwargs):
                return copier_type
        raise NotImplementedError(f"no valid copier for {source}")


class Copier(metaclass=ABCMeta):
    """ The base class for definition copiers """

    def __init__(self, source: str, **kwargs):
        self._source = source
        self._kwargs = {}

        for k, v in kwargs.items():
            if k in ["conflicts", "destination", "root_path"]:
                setattr(self, f"_{k}", v)
            else:
                self._kwargs[k] = v

        self._kwargs = kwargs

        if hasattr(self, "_conflicts"):
            if type(self._conflicts) is not list:
                raise ValueError("Conflicts must be a list of filenames to disallow")

    @abstractstaticmethod
    def type_match(source: str, **kwargs) -> bool:  # pragma: no cover
        """ type_match determins if the source is supported/handled by a copier """
        pass

    @abstractmethod
    def copy(self, **kwargs) -> None:  # pragma: no cover
        """ copy executes the copy from the source, into the working path """
        pass

    @property
    def root_path(self):
        """ root_path returns an optional root path to use for relative file operations """
        if hasattr(self, "_root_path"):
            return self._root_path
        else:
            return ""

    @property
    def conflicts(self):
        """ conflicts returns a list of disallowed files """
        if hasattr(self, "_conflicts"):
            return self._conflicts
        else:
            return []

    @property
    def source(self):
        """ source contains the source path providede """
        return self._source

    def get_destination(self, make_dir: bool = True, **kwargs) -> str:
        """ get_destination returns the destination path, and optionally makes the destinatination directory """
        if not (hasattr(self, "_destination") or "destination" in kwargs.keys()):
            raise ValueError("no destination provided")
        if "destination" in kwargs:
            d = kwargs["destination"]
        else:
            d = self._destination

        if make_dir:
            make_d = Path(d)
            make_d.mkdir(parents=True, exist_ok=True)

        return d

    def check_conflicts(self, path: str) -> None:
        """ Checks for files with conflicting names in a path """
        conflicting = []
        if self.conflicts:
            check_path = Path(path)
            for check_file in check_path.glob("*"):
                if check_file.name in self.conflicts:
                    conflicting.append(check_file.name)

        if conflicting:
            raise FileExistsError(f"{','.join(conflicting)}")


@CopyFactory.register("git")
class GitCopier(Copier):
    def copy(self, **kwargs) -> None:
        """ copy clones a remote git repo, and puts the requested files into the destination """
        dest = self.get_destination(**kwargs)
        branch = "master"

        if "branch" in kwargs:
            branch = kwargs["branch"]
        if "sub_path" in kwargs:
            sub_path = kwargs["sub_path"].strip("/")
        else:
            sub_path = ""

        self.make_temp()
        temp_path = f"{self._temp_dir}/{sub_path}"

        pipe_exec(
            f"git clone {self._source} --branch {branch} --single-branch ./",
            cwd=self._temp_dir,
        )

        try:
            self.check_conflicts(temp_path)
        except FileExistsError as e:
            self.clean_temp()
            raise e

        shutil.copytree(temp_path, dest, dirs_exist_ok=True)
        self.clean_temp()

    @staticmethod
    def type_match(source: str, **kwargs) -> bool:
        """ type matches uses git to see if the source is a valid git remote """
        try:
            (return_code, _, _) = pipe_exec(f"git ls-remote {source}")
        except (PermissionError, FileNotFoundError):
            return False
        if return_code == 0:
            return True
        return False

    def make_temp(self) -> None:
        if hasattr(self, "_temp_dir"):
            pass
        self._temp_dir = tempfile.mkdtemp()

    def clean_temp(self) -> None:
        """ clean_temp removes the temporary path used by this copier """
        if hasattr(self, "_temp_dir"):
            shutil.rmtree(self._temp_dir, ignore_errors=True)
            del self._temp_dir


@CopyFactory.register("fs")
class FileSystemCopier(Copier):
    def copy(self, **kwargs) -> None:
        """ copy copies files from a local source on the file system to a destination path """
        dest = self.get_destination(**kwargs)
        self.check_conflicts(self.local_path)
        shutil.copytree(self.local_path, dest, dirs_exist_ok=True)

    @property
    def local_path(self):
        """ local_path returns a complete local file system path """
        if not hasattr(self, "_local_path"):
            self._local_path = self.make_local_path(self.source, self._root_path)
        return self._local_path

    @staticmethod
    def type_match(source: str, **kwargs) -> bool:
        # check if there is a local file matching "source" if so, return true, if not return false
        if "root_path" in kwargs:
            source = FileSystemCopier.make_local_path(source, kwargs["root_path"])

        if os.path.isdir(source) or os.path.isfile(source):
            return True
        return False

    @staticmethod
    def make_local_path(source: str, root_path: str) -> str:
        """ make_local_path appends together known path objects to provide a local path """
        full_path = f"{root_path}/{source}"
        full_path = re.sub(r"/+", "/", full_path)
        return full_path



# Copyright 2020 Richard Maynard (richard.maynard@gmail.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import shlex
import subprocess


def pipe_exec(args, stdin=None, cwd=None, env=None):
    """
    A function to accept a list of commands and pipe them together.

    Takes optional stdin to give to the first item in the pipe chain.
    """
    count = 0  # int used to manage communication between processes
    commands = []  # listed used to hold all the popen objects

    # use the default environment if one is not specified
    if env is None:
        env = os.environ.copy()

    # if a single command was passed as a string, make it a list
    if not isinstance(args, list):
        args = [args]

    # setup various arguments for popen/popen.communicate, account for optional stdin
    popen_kwargs = {
        "stdout": subprocess.PIPE,
        "stderr": subprocess.PIPE,
        "cwd": cwd,
        "env": env,
    }
    popen_stdin_kwargs = {}
    communicate_kwargs = {}
    if stdin is not None:
        popen_stdin_kwargs["stdin"] = subprocess.PIPE
        communicate_kwargs["input"] = stdin.encode()

    # handle the first process
    i = args.pop(0)
    commands.append(
        subprocess.Popen(shlex.split(i), **popen_kwargs, **popen_stdin_kwargs)
    )

    # handle any additional arguments
    for i in args:
        popen_kwargs["stdin"] = commands[count].stdout
        commands.append(subprocess.Popen(shlex.split(i), **popen_kwargs))
        commands[count].stdout.close()
        count = count + 1

    # communicate with first command, ensure it gets any optional input
    commands[0].communicate(**communicate_kwargs)

    # communicate with final command, which will trigger the entire pipeline
    stdout, stderr = commands[-1].communicate()
    returncode = commands[-1].returncode

    return (returncode, stdout, stderr)



# Copyright 2020 Richard Maynard (richard.maynard@gmail.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import copy
import json
from typing import OrderedDict

import click
from tfworker import constants as const
from tfworker.util.copier import CopyFactory

TERRAFORM_TPL = """\
terraform {{
{0}{1}
}}
"""


class ReservedFileError(Exception):
    pass


class Definition:
    def __init__(
        self,
        definition,
        body,
        deployment,
        global_remote_vars,
        global_template_vars,
        global_terraform_vars,
        providers,
        repository_path,
        temp_dir,
        tf_version_major,
        limited=False,
    ):
        self.tag = definition
        self._body = body
        self._path = body.get("path")
        self._remote_vars = self.make_vars(
            body.get("remote_vars", collections.OrderedDict()), global_remote_vars
        )
        self._template_vars = self.make_vars(
            body.get("template_vars", collections.OrderedDict()), global_template_vars
        )
        self._terraform_vars = self.make_vars(
            body.get("terraform_vars", collections.OrderedDict()), global_terraform_vars
        )

        self._deployment = deployment
        self._repository_path = repository_path
        self._providers = providers
        self._temp_dir = temp_dir
        self._tf_version_major = tf_version_major
        self._limited = limited

    @property
    def body(self):
        return self._body

    @property
    def limited(self):
        return self._limited

    @property
    def path(self):
        return self._path

    def prep(self, backend):
        """ prepare the definitions for running """
        target = f"{self._temp_dir}/definitions/{self.tag}".replace("//", "/")

        try:
            c = CopyFactory.create(
                self.path,
                root_path=self._repository_path,
                conflicts=const.RESERVED_FILES,
            )
        except NotImplementedError:
            click.secho(
                f"could not handle source path {self.path} for definition {self.tag}, either file does not exist or could not handle remote URI",
                fg="red",
            )
            raise SystemExit(1)

        remote_options = dict(self.body.get("remote_path_options", {}))

        try:
            c.copy(destination=target, **remote_options)
        except FileExistsError as e:
            raise ReservedFileError(e)

        # Prepare variables
        self._template_vars["deployment"] = self._deployment
        self._terraform_vars["deployment"] = self._deployment

        # Create local vars from remote data sources
        if len(list(self._remote_vars.keys())) > 0:
            with open(f"{target}/worker-locals.tf", "w+") as tflocals:
                tflocals.write("locals {\n")
                for k, v in self._remote_vars.items():
                    tflocals.write(f"  {k} = data.terraform_remote_state.{v}\n")
                tflocals.write("}\n\n")

        # create remote data sources, and required providers
        remotes = list(map(lambda x: x.split(".")[0], self._remote_vars.values()))
        with open(f"{target}/terraform.tf", "w+") as tffile:
            tffile.write(f"{self._providers.hcl()}\n\n")
            required_providers = ""
            if self._tf_version_major >= 13:
                required_providers = f"\n\n{self._providers.required_providers()}"
            tffile.write(
                TERRAFORM_TPL.format(
                    f"{backend.hcl(self.tag)}",
                    required_providers,
                )
            )
            tffile.write(backend.data_hcl(remotes))

        # Create the variable definitions
        with open(f"{target}/worker.auto.tfvars", "w+") as varfile:
            for k, v in self._terraform_vars.items():
                varfile.write(f"{k} = {self.vars_typer(v)}\n")

    @staticmethod
    def quote_str(some_string):
        """Put literal quotes around a string."""
        return f'"{some_string}"'

    def make_vars(self, local_vars, global_vars):
        """Make a variables dictionary based on default vars, as well as specific vars for an item."""
        global_vars = global_vars or collections.OrderedDict()
        item_vars = copy.deepcopy(global_vars)
        for k, v in local_vars.items():
            item_vars[k] = v
        return item_vars

    @staticmethod
    def vars_typer(v, inner=False):
        """
        vars_typer is used to assemble variables as they are parsed from the yaml configuration
        into the required format to be used in terraform
        """
        if v is True:
            return "true"
        elif v is False:
            return "false"
        elif isinstance(v, list):
            rval = []
            for val in v:
                result = Definition.vars_typer(val, inner=True)
                try:
                    rval.append(result.strip('"').strip("'"))
                except AttributeError:
                    rval.append(result)
            if inner:
                return rval
            else:
                return json.dumps(rval)
        elif isinstance(v, OrderedDict) or isinstance(v, dict):
            rval = {}
            for k, val in v.items():
                result = Definition.vars_typer(val, inner=True)
                try:
                    rval[k] = result.strip('"').strip("'")
                except AttributeError:
                    rval[k] = result
            if inner:
                return rval
            else:
                return json.dumps(rval)
        return f'"{v}"'


class DefinitionsCollection(collections.abc.Mapping):
    def __init__(
        self,
        definitions,
        deployment,
        limit,
        plan_for,
        providers,
        repository_path,
        rootc,
        temp_dir,
        tf_version_major,
    ):
        self._body = definitions
        self._plan_for = plan_for
        self._definitions = collections.OrderedDict()
        self._limit = True if len(limit) > 0 else False
        self._limit_size = len(limit)
        for definition, body in definitions.items():
            self._definitions[definition] = Definition(
                definition,
                body,
                deployment,
                rootc.remote_vars_odict,
                rootc.template_vars_odict,
                rootc.terraform_vars_odict,
                providers,
                repository_path,
                temp_dir,
                tf_version_major,
                True if limit and definition in limit else False,
            )

    def __len__(self):
        return len(self._definitions)

    def __getitem__(self, value):
        if type(value) == int:
            return self._definitions[list(self._definitions.keys())[value]]
        return self._definitions[value]

    def __iter__(self):
        return self.iter(honor_destroy=True)

    def iter(self, honor_destroy=False):
        if honor_destroy:
            if self._plan_for == "destroy":
                return iter(reversed(list(self._definitions.values())))
        return iter(self._definitions.values())

    def limited(self):
        # handle the case where nothing is filtered
        iter_size = len(
            list(filter(lambda d: d.limited, self.iter(honor_destroy=True)))
        )
        if iter_size == 0:
            # a limit was supplied, but not matched, raise an error
            if self._limit:
                raise ValueError("no definitions matching --limit")
            # the run is not limited to anything, so return everything
            else:
                return self.iter(honor_destroy=True)
        elif iter_size < self._limit_size:
            # not all limit items are matched
            raise ValueError("not all definitions match --limit")
        else:
            return iter(filter(lambda d: d.limited, self.iter(honor_destroy=True)))

    @property
    def body(self):
        return self._body
